/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 04/06/22.
//

#ifndef FGGL_ECS3_PROTOTYPE_LOADER_HPP
#define FGGL_ECS3_PROTOTYPE_LOADER_HPP

#include "yaml-cpp/yaml.h"

#include "fggl/math/types.hpp"
#include "fggl/data/storage.hpp"
#include "fggl/data/model.hpp"
#include "fggl/data/procedural.hpp"

#include "fggl/ecs3/ecs.hpp"

namespace fggl::ecs3 {


	void load_prototype_node( ecs3::World& world, const YAML::Node& node) {
		auto prototype = world.create(true);
		auto& typeReg = world.types();

		// store the prototype's name for future use
		auto* nameComp = world.get<ecs3::EntityMeta>( prototype );
		nameComp->typeName = node["name"].as<std::string>();

		// grab the components and iterate them
		if ( node["components"] ) {
			world.createFromSpec( prototype, node["components"]);
		}
	}

	void load_prototype_file( ecs3::World& world, data::Storage& storage, const std::string& name ) {
		auto path = storage.resolvePath( data::Data, name );

		auto root = YAML::LoadFile( path.string().c_str() );
		for (const auto& node : root["prefabs"] ) {
			load_prototype_node(world, node);
		}
	}

} // namespace fggl::ecs3

namespace YAML {

	template<>
	struct convert<fggl::math::vec3> {
		static Node encode(const fggl::math::vec3& rhs){
			Node node;
			node.push_back(rhs.x);
			node.push_back(rhs.y);
			node.push_back(rhs.z);
			return node;
		}

		static bool decode(const Node& node, fggl::math::vec3& rhs) {
			if ( !node.IsSequence() || node.size() != 3) {
				return false;
			}

			rhs.x = node[0].as<float>();
			rhs.y = node[1].as<float>();
			rhs.z = node[2].as<float>();
			return true;
		}
	};

	template<>
	struct convert<fggl::math::quat> {
		static Node encode(const fggl::math::quat& rhs){
			Node node;
			node.push_back(rhs.x);
			node.push_back(rhs.y);
			node.push_back(rhs.z);
			node.push_back(rhs.w);
			return node;
		}

		static bool decode(const Node& node, fggl::math::quat& rhs) {
			if ( !node.IsSequence() || node.size() != 3) {
				return false;
			}

			rhs.x = node[0].as<float>();
			rhs.y = node[1].as<float>();
			rhs.z = node[2].as<float>();
			rhs.w = node[3].as<float>();
			return true;
		}
	};

	template<>
	struct convert<fggl::phys::BodyType> {
		const static std::string TYPE_KINEMATIC;
		const static std::string TYPE_STATIC;
		const static std::string TYPE_DYN;

		static Node encode(const fggl::phys::BodyType& rhs) {
			switch (rhs) {
				case fggl::phys::BodyType::KINEMATIC:
					return Node(TYPE_KINEMATIC);
				case fggl::phys::BodyType::STATIC:
					return Node(TYPE_STATIC);
				default:
				case fggl::phys::BodyType::DYNAMIC:
					return Node(TYPE_DYN);
			}
		}
		static bool decode(const Node& node, fggl::phys::BodyType& rhs) {
			const auto value = node.as<std::string>();
			if ( value == TYPE_KINEMATIC ) {
				rhs = fggl::phys::BodyType::KINEMATIC;
			} else if ( value == TYPE_STATIC ) {
				rhs = fggl::phys::BodyType::STATIC;
			} else {
				rhs = fggl::phys::BodyType::DYNAMIC;
			}
			return true;
		}
	};

	const std::string convert<fggl::phys::BodyType>::TYPE_KINEMATIC = "kinematic";
	const std::string convert<fggl::phys::BodyType>::TYPE_STATIC = "static";
	const std::string convert<fggl::phys::BodyType>::TYPE_DYN = "dynamic";

}

namespace fggl::ecs {

	constexpr int DEFAULT_STACKS = 16;
	constexpr int DEFAULT_SLICES = 16;
	constexpr const char* SHAPE_SPHERE_VALUE{"sphere"};
	constexpr const char* SHAPE_BOX_VALUE{"box"};

	static void process_mesh(data::Mesh& out, const YAML::Node& shape) {
		auto transform = data::OFFSET_NONE;
		if ( shape["offset"] ) {
			auto scaleFactor = shape["offset"].as<math::vec3>();
			transform = glm::translate(transform,  scaleFactor);
		}

		if ( shape["scale"] ) {
			auto scaleFactor = shape["scale"].as<math::vec3>();
			transform = glm::scale(transform,  scaleFactor);
		}

		// now the shape itself
		auto shapeType = shape["type"].as<std::string>();
		if ( shapeType == SHAPE_BOX_VALUE ) {
			data::make_cube(out, transform);
		} else if ( shapeType == SHAPE_SPHERE_VALUE ) {
			int stacks = shape["stacks"].as<int>(DEFAULT_STACKS);
			int slices = shape["slices"].as<int>(DEFAULT_SLICES);
			data::make_sphere(out, transform, stacks, slices);
		}
	}

}

#endif //FGGL_ECS3_PROTOTYPE_LOADER_HPP
